#! /usr/bin/env python3

"""Working with groups and users in vk"""

import sys
import json
import os
import time
from .vkreq import apply_vk_method, Executor
from ..utils import exception_handler


def _mkdir_rec(*dir_chain):
    """Recursive mkdir. For example, use `_mkdir_rec('a', 'b', 'c')` to create
    directory 'a/b/c'. Directories 'a' and 'b' will be created too if need"""
    path = ''
    for name in dir_chain:
        path += (name + os.path.sep)
        if not os.path.isdir(path):
            os.mkdir(path)
    return path


class VKObj():
    """Superclass for vk objects"""

    def __repr__(self):
        """Universal verbose view of class instance"""

        # class name
        repr_ = '[{}]\n'.format(self.__class__.__name__)

        # walk through attached fields of class instance
        d = self.__dict__
        keys = sorted(list(d.keys()))
        for key in keys:

            # field name
            repr_ += '{}: '.format(key).ljust(18)

            # empty case
            val = d[key]
            if not val:
                repr_ += '\n'
                continue

            # field value
            if isinstance(val, list) or isinstance(val, tuple):

                # items view
                if len(str(val[0]) + str(val[0])) < 30:
                    val_str = str(val)
                else:
                    val_str = '[{}]'.format(val[0].__class__.__name__)

                # items type, length
                val_str += ' len {}'.format(len(val))

                # if there are repetitions, output unique elements number
                try:
                    uniq_length = len(set(val))
                    if uniq_length != len(val):
                        val_str += ', uniq items {}'.format(uniq_length)
                except Exception:
                    pass

            elif isinstance(val, dict):
                # {int} len N
                random_item = next(iter(val.values()))
                val_str = '{{{}}} len {}'.format(
                    random_item.__class__.__name__,
                    len(val))

            else:
                # simple value
                val_str = '{}'.format(val)

            # shrink if need
            if len(val_str) > 52:
                val_str = '{}  ....  {}'.format(val_str[:22], val_str[-22:])

            repr_ += val_str + '\n'

        return repr_

    def dump(self, rewrite=False):
        """Save data about object to data/obj/../"""

        # add date to object
        date = time.strftime("%F-%H%M%S")
        self.dump_date = date

        # dump data
        dir_path = _mkdir_rec('data', 'obj', self.__class__.__name__)
        file_path = os.path.join(dir_path, self._get_key())
        if rewrite or not os.path.isfile(file_path):
            with open(file_path, 'w') as fp:
                json.dump(self.__dict__, fp)
        else:
            old_version = json.load(open(file_path))
            old_version.update(self.__dict__)
            with open(file_path, 'w') as fp:
                json.dump(old_version, fp)

    def dump_to_frozen(self):
        """Save data about object to data/frozen_obj/../"""

        # add date to object
        date = time.strftime("%F-%H%M%S")
        self.dump_date = date

        # dump data
        dir_path = _mkdir_rec('data', 'frozen_obj', self.__class__.__name__)
        date = '_' + date
        with open(os.path.join(dir_path, self._get_key() + date), 'w') as fp:
            json.dump(self.__dict__, fp)

    def open(self):
        """Restore data of object from data/obj/../"""

        # get path to object
        path = os.path.join('data', 'obj', self.__class__.__name__,
                            self._get_key())
        if not os.path.isfile(path):
            return

        # read object
        with open(path) as fp:
            obj = json.load(fp)
        for field in obj:
            setattr(self, field, obj[field])

    def open_last_frozen(self):
        """Restore recent data of object from data/frozen_obj/../"""

        # get path to recent object
        dir_path = os.path.join('data', 'frozen_obj', self.__class__.__name__)
        if not os.path.isdir(dir_path):
            return
        files = next(os.walk(dir_path))[2]
        files = [x for x in files if x.startswith(self._get_key() + '_')]
        if not files:
            return
        obj_path = os.path.join(dir_path, sorted(files)[-1])

        # read object
        with open(obj_path) as fp:
            obj = json.load(fp)
        for field in obj:
            setattr(self, field, obj[field])


# Check user for existance university of univer_ids in list of his universities
def _is_student(user_item, univer_ids):
    """Does the user profile indicate that he studied at a university?"""

    # field universities
    if 'universities' in user_item:
        for i in user_item['universities']:
            if i['id'] in univer_ids:
                return True

    # field occupation
    if 'occupation' in user_item:
        occ = user_item['occupation']
        if occ['type'] == 'university':
            if occ['id'] in univer_ids:
                return True

    return False


class Group(VKObj):
    """A class containing information about the community and a list of
    its members.

    The easiest and fastest way to upload information - use `Group.load()`.
    For fast loading of many small communities use outclass function
    `load_groups`."""

    def __init__(self, group_id):

        if not str(group_id).isdigit():
            print('Wrong group_id {}. It should be digital'.format(group_id))
            return

        # Init
        self.group_id = str(group_id)
        self.count = 0
        self.members = []
        self.cumul_members = []  # cumulative list of former & current members
        self.univers_data = {}

    def _get_key(self):
        """Key for dump"""
        return self.group_id

    def load(self, extra_getById=()):
        """Load from vk.com initial community information
        and full list of community member's ids"""

        # load phase 0
        e = Executor()
        self.load_ph0_fill_requests(e, extra_getById)
        for i in range(1000, 24000, 1000):
            e.add_request('groups.getMembers',
                          self.load_ph1_parse,
                          group_id=self.group_id, offset=i, count=1000)
        e.emit_requests()

        # load phase 1
        if self.count > 24000:
            for i in range(24000, self.count, 1000):
                e.add_request('groups.getMembers',
                              self.load_ph1_parse,
                              group_id=self.group_id, offset=i, count=1000)
            e.emit_requests()

        # update cumulative data
        self.update_cumulative()

    def update_cumulative(self):
        """If we read the data locally (`Group.open()`), and then received
        the current data from the server (`Group.load()` or `load_groups`),
        we combine the new data with the old ones and save in special fields
        of \"cumulative\" data."""
        self.cumul_members = list(set(self.cumul_members + self.members))

    def load_ph0_fill_requests(self, executor, extra_getById=()):
        """Add to executor 1 request groups.getById and 1 request
        groups.getMembers for getting initial vk community information and
        ids of first 1000 members.

        `executor` - object of class `Executor`
        `extra_getById` - additional fields you can specify
                          for vk API method getById"""

        # temporarily save, Group.load_ph0_parse will remove it
        self.extra_getById = extra_getById

        fields_s = ','.join(('description',) + extra_getById)
        executor.add_request('groups.getById',
                             self.load_ph0_parse,
                             group_id=self.group_id, fields=fields_s)
        executor.add_request('groups.getMembers',
                             self.load_ph0_parse,
                             group_id=self.group_id, offset=0, count=1000)

    def load_ph0_parse(self, response):
        """Processor for response of `Group.load_ph0_fill_requests()`"""

        # check and prepare response to read
        if not response:
            return
        if isinstance(response, list):
            response = response[0]

        # remove temporary fields list
        if hasattr(self, 'extra_getById'):
            extra_getById = self.extra_getById
            del self.extra_getById
        else:
            extra_getById = ()

        # save data
        for field in ('screen_name', 'count', 'name', 'description',
                      'photo_200') + extra_getById:
            if field in response:
                setattr(self, field, response[field])
        if 'items' in response:
            self.members = response['items']

    def load_ph1_fill_requests(self, executor):
        """Add to executor requests groups.getMembers for getting
        full list of community member's ids"""

        if not hasattr(self, 'count'):
            return
        for i in range(1000, self.count, 1000):
            executor.add_request('groups.getMembers',
                                 self.load_ph1_parse,
                                 group_id=self.group_id, offset=i, count=1000)

    def load_ph1_parse(self, response):
        """Processor for response of `Group.load_ph1_fill_requests()`"""

        if not response:
            return

        # save data
        if 'items' in response:
            self.members += response['items']

    def load_students(self, university):
        """If you have already executed `Group.load()` you can to get
        a separate list of students in this group.
        Result will be saved in `Group.univers_data`"""

        # make auxiliary class instance add put into it old data if they are
        univer = _UniversityInGroup(university.ids)
        if university.name in self.univers_data:
            for f in univer.__dict__.keys():
                if f != 'students':
                    setattr(univer, f, self.univers_data[university.name][f])

        # load
        e = Executor()
        univer.load_students_fill_requests(e, self)
        e.emit_requests()

        # update cumulative data
        univer.cumul_students = list(set(univer.cumul_students
                                         + univer.students))

        # calc share of students in community
        univer.univer_fraction = len(univer.students)/len(self.members)

        # Save data
        self.univers_data[university.name] = {}
        for f in univer.__dict__.keys():
            self.univers_data[university.name][f] = getattr(univer, f)

    def __repr__(self):

        # main information
        repr_ = VKObj.__repr__(self)

        # universities information
        for uname in self.univers_data:

            # make auxiliary class instance and fill it
            univer = _UniversityInGroup()
            for f in self.univers_data[uname]:
                setattr(univer, f, self.univers_data[uname][f])
            setattr(univer, 'name', uname)

            # get usual class representation and format it a bit
            lines = VKObj.__repr__(univer)
            lines = lines[:-1].replace('\n', '\n    ')
            lines = lines.replace('[_UniversityInGroup]', '[UniverData]')

            # add to class instance representation
            repr_ += '\n    {}\n'.format(lines)

        return repr_

    def is_empty(self):
        return self.count == 0


class _UniversityInGroup():
    """Auxiliary class for `Group.load_students()` and `Group.__repr__()`"""

    def __init__(self, university_ids=[]):
        self.univer_ids = university_ids
        self.students = []
        self.cumul_students = []
        self.univer_fraction = 0.

    def load_students_fill_requests(self, executor, group):
        """Add to executor request for `Group.load_students()`"""

        # Limiting the server to the length of the request does not allow
        # making a step of 1000 (the maximum allowed by api) requests.
        # But it does not matter: the critical limitation of method users.get
        # is it's runtime on the server.
        step = 100

        for i in range(0, group.count, step):
            user_ids_s = ','.join(map(str, group.members[i:i+step]))
            executor.add_request('users.get',
                                 self.load_students_parse,
                                 user_ids=user_ids_s,
                                 fields='universities,occupation')

    def load_students_parse(self, response):
        """Process to executor request for `Group.load_students()`"""

        if not response:
            return

        # save data
        usr_objs = [x for x in response if _is_student(x, self.univer_ids)]
        self.students += [x['id'] for x in usr_objs]


def load_groups(groups, extra_getById=()):
    """Fast load groups data thanks to vk api method `execute`.
    `groups` - list of objects of `Group` class"""

    e = Executor()

    # load phase 1
    for g in groups:
        g.load_ph0_fill_requests(e, extra_getById)
    e.emit_requests()

    # load phase 2
    for g in groups:
        g.load_ph1_fill_requests(e)
    e.emit_requests()

    # update cumulative data
    for g in groups:
        g.update_cumulative()


class University(VKObj):
    """A class containing information about the university"""

    def __init__(self, name, ids=[]):

        try:
            # check name
            if not name.islower():
                raise
            for ch in name:
                if ch not in 'abcdefghijklmnopqrstuvwxyz_0123456789':
                    raise

            # init data
            self.name = name
            self.ids = list(map(int, ids))

        except Exception as e:
            exception_handler(e, 'Invalid character in university name: '
                              + '"{}"'.format(ch))

    def _get_key(self):
        """Key for dump"""
        return self.name


class User(VKObj):
    """TODO: rewrite like Group"""

    def __init__(self, user_id, univer_ids=[]):

        # Assert
        if type(user_id) != int:
            print('Wrong type of user id: ' + str(user_id)
                  + ' (must be \'str\')')
            sys.exit()

        # Init
        self.user_id = user_id
        self.count = 0
        self.friends = []
        # for univer friends:
        self.univer_ids = univer_ids
        self.univer_friends = []  # university membership is listed in account
        self.univer_degree = 0.  # len(univer_friends)/len(friends)
        # other data:
        self.is_student = False
        self.photo_max = ''
        self.first_name = ''
        self.last_name = ''
        self.nickname = ''
        self.scrname = ''
        self.byear = 0  # year of birthday
        self.sex = 0  # like in vk API: 0 - unknown, 1 - female, 2 - male
        # flag of relevance
        self.is_relevant = False

    def load_info(self):

        # Getting main info about user
        response = apply_vk_method('users.get',
                                   user_id=str(self.user_id),
                                   fields='universities,occupation,photo_max,'
                                          + 'screen_name,nickname,bdate,sex')
        if response:
            self.is_student = _is_student(response['response'],
                                          self.univer_ids)
            if response['response']['photo_max']:
                self.photo_max = response['response']['photo_max']
            else:
                self.photo_max = 'https://vk.com/images/camera_200.png?ava=1'
            self.first_name = response['response']['first_name']
            self.last_name = response['response']['last_name']
            if response['response']['nickname']:
                self.nickname = response['response']['nickname']
            self.scrname = response['response']['scrname']
            if response['response']['bdate']:
                bdate = response['response']['bdate']
                if len(bdate.split('.')) == 3:
                    self.byear = int(bdate.split('.')[2])
            self.sex = int(response['response']['sex'])
        else:
            return

        # Getting count and first 5000 friends
        response = apply_vk_method('friends.get', user_id=str(self.user_id),
                                   offset=0, count=5000,
                                   fields='universities,occupation')
        if response:
            self.count = response['response']['count']
            users = response['response']['items']
        else:
            return

        # Getting the remaining friends
        for i in range(1, int((self.count-1)/5000) + 1):
            response = apply_vk_method('friends.get',
                                       user_id=str(self.user_id),
                                       offset=i*5000, count=5000,
                                       fields='universities,occupation')
            if bool(response):
                users += response['response']['items']

        self.friends = [x['id'] for x in users]
        self.univer_friends = \
            [x['id'] for x in users if _is_student(x, self.univer_ids)]
        try:
            self.univer_degree = len(self.univer_friends) / len(self.friends)
        except ZeroDivisionError:
            self.univer_degree = 0
        self.is_relevant = True

        # During the download, the composition of the group could change
        self.count = len(self.friends)

    def get_json_repr(self):
        return {
            'user_id': self.user_id,
            'count': self.count,
            'friends': self.friends,
            'univer_ids': self.univer_ids,
            'univer_friends': self.univer_friends,
            'univer_degree': self.univer_degree,
            'is_student': self.is_student,
            'photo_max': self.photo_max,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'nickname': self.nickname,
            'scrname': self.scrname,
            'byear': self.byear,
            'sex': self.sex,
            'is_relevant': self.is_relevant}

    def save_info(self):
        if not os.path.isdir('data/users'):
            os.mkdir('data/users')
        with open('data/users/' + str(self.user_id), 'w') as f:
            json.dump(self.get_json_repr(), f)

    def is_info(self):
        return os.path.isfile('data/users/' + str(self.user_id))

    def read_info(self):
        if not self.is_info():
            print('Info about user ' + str(self.user_id) + ' is absent')
            raise
        with open('data/users/' + str(self.user_id), 'r') as f:
            json_repr = json.load(f)
        self.user_id = json_repr['user_id']
        self.count = json_repr['count']
        self.friends = json_repr['friends']
        self.univer_ids = json_repr['univer_ids']
        self.univer_friends = json_repr['univer_friends']
        self.univer_degree = json_repr['univer_degree']
        self.is_student = json_repr['is_student']
        self.photo_max = json_repr['photo_max']
        self.first_name = json_repr['first_name']
        self.last_name = json_repr['last_name']
        self.nickname = json_repr['nickname']
        self.scrname = json_repr['scrname']
        self.byear = json_repr['byear']
        self.sex = json_repr['sex']
        self.is_relevant = json_repr['is_relevant']

    def load_and_save_info_if_absent(self):
        if not self.is_info():
            self.load_info()
            self.save_info()
        else:
            self.read_info()
            if not self.is_relevant:
                self.load_info()
                self.save_info()

    def is_empty(self):
        return self.count == 0



#! /usr/bin/env python3

"""The most common interfaces"""

import sys
import traceback
import __main__


def is_py_interactive():
    return not hasattr(__main__, '__file__')


def exception_handler(e, msg=None):

    # print stack
    if not isinstance(e, SystemExit):
        traceback.print_exc()

    # print message if need
    if msg:
        print('\n{}\n'.format(msg))

    # exit if not interpreter mode enabled
    if (not isinstance(e, SystemExit)
            and not is_py_interactive()):
        sys.exit()



#! /usr/bin/env python3

"""Implementation of the main functions of the application: editing user data,
thematic search, other console commands"""

import sys
import re
from collections import Counter
from .report import Report
from . import vklib as vk
from .vklib import apply_vk_method
from .usrdata import UsrData
from .utils import exception_handler


####################################################################
##                        Account commands                        ##
####################################################################


def add_account():
    """Interactive function of account adding"""

    try:
        # Read data about account

        # Account type
        in_str = input('What type of account do you want to create?\n'
                       + 'Input please one letter - [e]mail, [v]k.ru\n'
                       + 'or [t]elegram: ').lower()
        ac_type = {'e': 'email', 'v': 'vk', 't': 'telegram'}[in_str]

        # Account name
        ac_name = input('\nInput name of new account: ').lower()

        # Privacy
        print('\nAre your account fake (password will be stored\n'
              + 'unencrypted on your computer) or private (you\n'
              + 'will be forced to enter the password every run)?')
        in_str = input('Input please [f]ake or [p]rivate: ').lower()
        ac_privacy = {'f': 'fake', 'p': 'private'}[in_str]

        # User name
        if ac_type == 'email' or ac_type == 'vk':
            ac_user_name = input('\nInput email address: ')
        elif ac_type == 'telegram':
            ac_user_name = input('\nInput nickname: ').lstrip('@')
        else:
            raise

        # Password (it's fake account, so we don't use smth. like getpass)
        if ac_privacy == 'fake':
            ac_password = input('\nInput password: ')
        else:
            ac_password = None

        # Save account data
        acc_obj = {'uname': ac_user_name,
                   'password': ac_password,
                   'token': None,
                   'is_activated': False}
        UsrData().set(acc_obj, 'acc', ac_type, ac_name, correct_is_act=True)

    except Exception as e:
        exception_handler(e, 'Failed to enter account data')


def delete_account(ac_type, ac_name):
    """Deletion account ac_name of type ac_type from registry"""

    UsrData().del_('acc', ac_type, ac_name, correct_is_act=True)


def activate_account(ac_type, ac_name):
    """Choose active account"""

    u = UsrData()
    if ac_name in u.get('acc', ac_type):
        u.drop_activations('acc', ac_type)
        u.set(True, 'acc', ac_type, ac_name, 'is_activated')


def display_accounts():
    """Display all accounts and active marks"""

    accs = UsrData().get('acc')
    for ac_type in accs:

        # check existance
        if not accs[ac_type]:
            continue

        # print accounts of type ac_type
        print(ac_type)
        for ac_name in accs[ac_type]:
            acc_obj = accs[ac_type][ac_name]
            print('    {}: \t[{}|{}]{}'.format(ac_name,
                                               acc_obj['uname'],
                                               acc_obj['password'],
                                               ('    \t<- activated'
                                                if acc_obj['is_activated']
                                                else '')))
        print('')


####################################################################
##                      University commands                       ##
####################################################################


# Search hot university ids by un_groups
def search_hot_university_ids(un_groups, un_ids):

    # load members of groups
    groups = [vk.Group(group_id['id']) for group_id in un_groups]
    vk.load_groups(groups)

    # unite groups members in a single users list
    users = []
    for group in groups:
        users += group.members
    users = list(set(users))

    # load education info about users
    step = 100
    e = vk.Executor()
    users_education = []
    for i in range(0, len(users), step):
        ids_str = ','.join(list(map(str, users[i:i+step])))
        e.add_request('users.get', user_ids=ids_str,
                      fields='occupation,education,universities',
                      processor=users_education.extend)
    e.emit_requests()

    # Count university appearances
    un_ids_cnt = {un_id: 0 for un_id in un_ids}
    for item in users_education:
        if 'university' in item:
            un_id = str(item['university'])
            if un_id in un_ids_cnt:
                un_ids_cnt[un_id] += 1
                continue
        if ('occupation' in item
                and item['occupation']['type'] == 'university'):
            un_id = str(item['occupation']['id'])
            if un_id in un_ids_cnt:
                un_ids_cnt[un_id] += 1
                continue

    # Filter out unused ids
    hot_ids = list(un_ids_cnt.items())
    hot_ids = [x for x in hot_ids if x[1] >= 10]
    hot_ids.sort(key=(lambda x: x[1]), reverse=True)
    hot_ids = [{'id': x[0], 'temp': x[1]} for x in hot_ids]

    return hot_ids


def add_university():
    """Interactive function of adding data abaot university"""

    # Read data about university
    try:

        # Name of university
        un_name = input(
            'Input please simple name of university (better in short latin;\n'
            + 'for example: mipt, msu, mgimo): \n> ')

        # Titles of university
        print('Input please all titles of university (for example:\n'
              + 'МФТИ, Физтех, MIPT). After every title press [ENTER].\n'
              + 'Finish input by [ENTER].')
        un_titles = []
        while True:
            title = input('> ')
            if not title:
                break
            un_titles.append(title)
        assert un_titles

        # Looking for a university in the vk database
        e = vk.Executor()
        for title in un_titles:
            e.add_request('database.getUniversities', q=title)
        e.emit_requests()
        un_items = []
        for r in e.responses:
            un_items.extend(r['items'])

        # Clarify the results by checking exact words match
        tmp = un_items
        un_items = []
        unique_ids = set()
        for item in tmp:

            # ensuring uniqueness of search results
            if item['id'] in unique_ids:
                continue
            unique_ids.add(item['id'])

            # check title existence
            for title in un_titles:
                for word in title.split():
                    if not re.search(r'\b' + word.lower() + r'\b',
                                     item['title'].lower()):
                        break
                else:
                    un_items.append(item)
                    break

        del tmp, unique_ids
        assert un_items

        # Ask the user to clarify the results
        print(
            'Database search results may contain extra items.\n'
            + 'Check it please. You will see chanks of 10 items in turn.\n'
            + 'For every chank enter the numbers corresponding to the wrong\n'
            + 'elements (for example, 239 if corr. three items are wrong).\n'
            + 'Then press [ENTER].')
        tmp = un_items
        un_items = []
        for i in range(0, len(tmp), 10):
            chunk = tmp[i:i+10]
            for j, item in enumerate(chunk):
                print('{}> {}'.format(j, item['title']))
            in_str = input('Wrong numbers: ')
            extra_indexes_list = [int(c) for c in in_str if c.isdigit()]
            extra_indexes_list.sort(reverse=True)
            for j in extra_indexes_list:
                del chunk[j]
            un_items.extend(chunk)
        del tmp
        assert un_items

        # Read very big groups which consists many students
        print('Input please id or domains of very very big groups\n'
              + 'which consists many students of added university.\n'
              + 'After every id press [ENTER]. Finish input by [ENTER].')
        un_big_groups = []
        while True:
            gr_id = input('> ')
            if not gr_id:
                break
            un_big_groups.append(gr_id)
        un_big_groups = vk.resolve_group_ids(un_big_groups)
        un_big_groups = [{'id': x[0], 'domain': x[1]} for x in un_big_groups]
        assert un_big_groups

        # Read groups which consists almost exclusively of students
        print('Input please id or domains of groups which consists\n'
              + 'almost exclusively of students of added university.\n'
              + 'After every id press [ENTER]. Finish input by [ENTER].')
        un_groups = []
        while True:
            gr_id = input('> ')
            if not gr_id:
                break
            un_groups.append(gr_id)
        un_groups = vk.resolve_group_ids(un_groups)
        un_groups = [{'id': x[0], 'domain': x[1]} for x in un_groups]

        # Read users whose almost all friends are students
        print('Input please id or domains of users whose almost all\n'
              + 'friends are students of added university.\n'
              + 'After every id press [ENTER]. Finish input by [ENTER].')
        un_users = []
        while True:
            us_id = input('> ')
            if not us_id:
                break
            us_id, us_scr = vk.resolve_user_ids(us_id)
            un_users.append({'id': us_id, 'domain': us_scr})

        # Search hot university ids by un_big_groups (unfortunately,
        # vk search method is not stable, so we use this approach)
        print('Start analysis of the prevalence of university identifiers.\n'
              + '(It could take several minutes)')
        un_ids = list(map(str, [x['id'] for x in un_items]))
        hot_ids = search_hot_university_ids(un_big_groups, un_ids)

        # Save university data
        univ_obj = {'titles': un_titles,
                    'big_groups': un_big_groups,
                    'crystal_groups': un_groups,
                    'crystal_users': un_users,
                    'all_ids': [x['id'] for x in un_items],
                    'hot_ids': hot_ids,
                    'is_activated': False}
        UsrData().set(univ_obj, 'univ', un_name, correct_is_act=True)

    except Exception as e:
        exception_handler(e, 'Failed to enter or add university data')


def delete_university(un_name):
    """Deletion un_name from registry"""

    UsrData().del_('univ', un_name, correct_is_act=True)


def activate_university(un_name):
    """Choose active university for analysis"""

    u = UsrData()
    if un_name in u.get('univ'):
        u.drop_activations('univ')
        u.set(True, 'univ', un_name, 'is_activated')


def display_universities():
    """Display data for all universities"""

    # Read universities registry
    univs = UsrData().get('univ')

    # Print
    for un_name in univs:
        univ_obj = univs[un_name]
        print(un_name + '{}'.format('\t\t\t\t<- ACTIVATED'
                                    if univ_obj['is_activated']
                                    else ''))
        print('Title:             {}'.format(', '.join(univ_obj['titles'])))
        s = ', '.join([x['domain'] for x in univ_obj['big_groups']])
        print('Big vk groups:     {}'.format(s))
        s = ', '.join([x['domain'] for x in univ_obj['crystal_groups']])
        print('Crystal vk groups: {}'.format(s))
        s = ', '.join([x['domain'] for x in univ_obj['crystal_users']])
        print('Crystal vk users:  {}'.format(s))

        # Identifier list with carry and indents
        print('VK ids:            ', end='')
        ids = [str(x) for x in univ_obj['all_ids']]
        ids2 = []
        for i in range(0, len(ids), 5):
            ids2.append(', '.join(ids[i:i+5]))
        s = ',\n                   '.join(x for x in ids2)
        print(s)

        # Hot identifiers with temperature in parentheses
        sum_cnt = sum([x['temp'] for x in univ_obj['hot_ids']])
        s = ', '.join(['{} ({:.1f} %)'.format(x['id'], 100*x['temp']/sum_cnt)
                      for x in univ_obj['hot_ids']])
        print('Hot VK ids:        {}'.format(s))

        print('')


####################################################################
##                        VK API commands                         ##
####################################################################


# Any user-defined vk API method
def vk_method(method_name, args):
    args_list = [x.split('=', 1) for x in args]
    args_dict = {x[0]: x[1] for x in args_list}
    return apply_vk_method(method_name, handle_api_errors=False, **args_dict)


####################################################################
##                         Other commands                         ##
####################################################################


# Search new crystal students among friends of given set of crystal students
def load_crystal_students_from_friends(given_crystal,
                                       exceptions_list,  # users not to be
                                                         # searched for
                                       univer_ids):
    new_crystal = []
    for user_id in given_crystal:
        u = vk.User(user_id, univer_ids)
        u.load_and_save_info_if_absent()
        u.read_info()
        new_crystal += u.univer_friends

    print('new_crystal: ' + str(len(new_crystal)))
    new_crystal = list(set(new_crystal))
    print('unique new_crystal: ' + str(len(new_crystal)))
    new_crystal = list(set(new_crystal) - set(exceptions_list))
    print('with except new_crystal: ' + str(len(new_crystal)))
    return new_crystal


# Load friends of given users list with counters of repetitions
# Return: generate dict {id0:num_of_repeat(id0), id1:..., ...}
def load_friends_of_users_with_cnt(users_list):

    friends_list = []
    for user_id in users_list:
        u = vk.User(user_id)
        u.read_info()
        friends_list += u.friends

    # Count repeating elements in list
    # (generate dict {id0:num_of_repeat(id0), id1:..., ...})
    friends_cnt = dict(Counter(friends_list))

    # Exclude original users
    users_set = set(users_list)
    friends_cnt_filtered = \
        {x: friends_cnt[x] for x in friends_cnt if x not in users_set}

    return friends_cnt_filtered


# Load ids of friends of users and write them to data/users_packs/pack_name
def load_users_friends_to_users_pack(user_ids, pack_name, univer_ids):

    friends_list = []
    for user_id in user_ids:
        u = vk.User(user_id, univer_ids)
        u.load_and_save_info_if_absent()
        u.read_info()
        friends_list += u.friends

    # Delete repeats and sort
    friends_list = list(set(friends_list))
    friends_list.sort()

    # Write
    vk.write_users_pack(friends_list, pack_name)


# Wrapper for method users.search
def users_search_wrapper(university_id, **p):
    response = apply_vk_method('users.search', university=university_id,
                               offset=0, count=1000, **p)
    if bool(response):
        return [x['id'] for x in response['response']['items']]
    else:
        return []


# Load list of phystechs using method users.search
# WARNING: Don't use it too often. Otherwise account will be banned.
# TODO: Обобщить на другие вузы (МГИМО, СПбГУ, Бауманка, МГЮА, универсальный),
#       хранить в спец. файле в data список вузов с id и ключевыми словами
def load_phystechs_by_search(pack_name):

    # Getting list of phystechs
    phystechs_list = users_search_wrapper(55111, {'sort': 0})
# TODO: После отладки новой версии раскоментить и удалить след. строку
    phystechs_list = phystechs_list[10]
#    phystechs_list += users_search_wrapper(55111, **{'sort': 1})
#    phystechs_list += users_search_wrapper(55111, **{'age_to': 25})
#    phystechs_list += users_search_wrapper(55111, **{'age_from': 26})
#    phystechs_list += users_search_wrapper(55111, **{'city': 1}) # Москва
#    #print('55111: ' + str(len(list(set(phystechs_list)))))
#    phystechs_list += users_search_wrapper(297, **{'sort': 0, 'sex': 1})
#    phystechs_list += users_search_wrapper(297, **{'sort': 1, 'sex': 1})
#    #print('sex 1: ' + str(len(list(set(phystechs_list)))))
#    phystechs_list += users_search_wrapper(297, **{'sort': 0, 'sex': 2})
#    phystechs_list += users_search_wrapper(297, **{'sort': 1, 'sex': 2})
#    #print('sex 2: ' + str(len(list(set(phystechs_list)))))
#    phystechs_list += users_search_wrapper(297, **{'age_to': 18})
#    for i in range(19, 36):
#        phystechs_list += users_search_wrapper(297, **{'age_from': i, 'age_to': i})
#    phystechs_list += users_search_wrapper(297, **{'age_from': 36, 'age_to': 37})
#    phystechs_list += users_search_wrapper(297, **{'age_from': 38, 'age_to': 40})
#    phystechs_list += users_search_wrapper(297, **{'age_from': 41, 'age_to': 45})
#    phystechs_list += users_search_wrapper(297, **{'age_from': 46, 'age_to': 50})
#    phystechs_list += users_search_wrapper(297, **{'age_from': 51, 'age_to': 55})
#    phystechs_list += users_search_wrapper(297, **{'age_from': 56, 'age_to': 60})
#    phystechs_list += users_search_wrapper(297, **{'age_from': 61, 'age_to': 70})
#    phystechs_list += users_search_wrapper(297, **{'age_from': 71})
#    #print('ages: ' + str(len(list(set(phystechs_list)))))
#    phystechs_list += users_search_wrapper(297, **{'city': 1}) # Москва
#    #print('Moscow: ' + str(len(list(set(phystechs_list)))))
#    phystechs_list += users_search_wrapper(297, **{'city': 857}) # Долгопрудный
#    #print('Dolgoprudny: ' + str(len(list(set(phystechs_list)))))
#    for n in ['Александр', 'Сергей', 'Дмитрий', 'Андрей', 'Алексей',
#              'Владимир', 'Михаил', 'Игорь', 'Евгений', 'Юрий', 'Никита',
#              'Олег', 'Николай', 'Иван', 'Павел']:
#        phystechs_list += users_search_wrapper(297,
#                                               **{'q': urllib.parse.quote(n)})
#    #print('Names: ' + str(len(list(set(phystechs_list)))))

    # Sort result & delete repeats
    phystechs_list = list(set(phystechs_list))
    phystechs_list.sort()

    # Write & return
    vk.write_users_pack(phystechs_list, pack_name)
    return


# Read from groups of set members of university (field 'univer_members')
def read_crystal_students_of_groups_pack(pack_name):

    groups = vk.read_groups_pack(pack_name)

    crystal_students = []
    for group_id in groups:
        g = vk.Group(group_id)
        g.open()
        crystal_students += g.univer_members

    return crystal_students


#
def make_users_list_with_add_info(users_list,
                                  univer_groups_packs_list,
                                  thematic_groups_packs_list):
    ext_users_list = [{'id': x,
                       'univer_groups': [],
                       'thematic_groups': []} for x in users_list]

    for groups_pack in univer_groups_packs_list:
        groups_list = vk.read_groups_pack(groups_pack)
        for group_id in groups_list:
            g = vk.Group(group_id)
            g.open()
            members_list = g.members
            for ext_user in ext_users_list:
                if ext_user['id'] in members_list:
                    ext_user['univer_groups'].append([group_id, g.name])

    for groups_pack in thematic_groups_packs_list:
        groups_list = vk.read_groups_pack(groups_pack)
        for group_id in groups_list:
            g = vk.Group(group_id)
            g.open()
            members_list = g.members
            for ext_user in ext_users_list:
                if ext_user['id'] in members_list:
                    ext_user['thematic_groups'].append([group_id, g.name])

    return ext_users_list


# Create report about thematic student by list with
# information (id, their groups packs) about them
def make_thematic_report(ext_users_list, univer_ids):

    # Begin report
    r = Report('thematic_students_list')
    i = 0
    r.add_str('<table>\n')

    for ext_user in ext_users_list:

        # Read user info
        u = vk.User(ext_user['id'], univer_ids)
        u.load_and_save_info_if_absent()
        u.read_info()

        # Write user description to report
        i += 1
        r.add_str('<tr>\n')
        r.add_str(' <td valign="top">' + str(i) + '.</td>\n')
        r.add_str(' <td><a href="https://vk.com/id' + str(ext_user['id'])
                  + '" target="_blank"><img height=50 src="' + u.photo_max
                  + '"></a></td>\n')
        r.add_str(' <td valign="top">\n')
        r.add_line('  ' + u.first_name + ' ' + u.last_name)
        r.add_str(' <td valign="top">\n')
        r.add_str('  MIPT: ')
        if u.is_student:
            r.add_str('PROFILE, ')
        for group in ext_user['univer_groups']:
            r.add_str('<a = href="https://vk.com/club' + group[0]
                      + '" target="_blank">' + group[1] + '</a>, ')
        r.add_str('<br>')
        r.add_str('  Thematic: ')
        for group in ext_user['thematic_groups']:
            r.add_str('<a = href="https://vk.com/club' + group[0]
                      + '" target="_blank">' + group[1] + '</a>, ')
        r.add_str(' </td>\n')
        r.add_str('</tr>\n')

    # Conclude report
    r.add_str('</table>\n')
    if not r.is_empty():
        r.conclude()
        r.dump()


def get_thematic_students_by_their_groups():

    univer_ids = [297, 55111]

    # Load members of mipt groups
    #vk.load_users_pack_by_groups_pack('mipt_crystal_gr', univer_ids)
    #vk.load_users_pack_by_groups_pack('mipt_very', univer_ids)
    #vk.load_users_pack_by_groups_pack('mipt_roughly', univer_ids)

    # Load friends of user 23681294 (Концертный Физтеха)
    # to data/users_packs/mipt_crystal_us
    #load_users_friends_to_users_pack(['23681294'], 'mipt_crystal_us',
    #                                 univer_ids)
    ## TODO: Фильтрации здесь или в конце
    ## TODO: убрать всё crystal в одну ф-ию
    ## TODO: раскоментить search
    ## TODO: clear в users и groups если нет в pack-ах

    # Search phystechs by and dump to data/users_packs/mipt_crystal_search
    # WARNING: too frequent calls to this function result in
    #          the method users.search being blocked
    #load_phystechs_by_search('mipt_crystal_search')
    ## TODO: Всем, кто найден поиском нужно проставлять is_student

    # Read phystechs from crystal sources
    #mipt_crystal_1_gusm = vk.read_users_pack('mipt_crystal_gr')
    #print('+ mipt_crystal_gr: ' + str(len(list(set(mipt_crystal_1_gusm)))))
    #mipt_crystal_1_gusm += vk.read_users_pack('mipt_crystal_us')
    #print('+ mipt_crystal_us: ' + str(len(list(set(mipt_crystal_1_gusm)))))
    #mipt_crystal_1_gusm += vk.read_users_pack('mipt_crystal_search')
    #print('+ mipt_crystal_search: ' + str(len(list(set(mipt_crystal_1_gusm)))))
    #mipt_crystal_1_gusm = list(set(mipt_crystal_1_gusm))

    # Search crystal phystechs in mipt-groups
    #mipt_crystal_1_gusm += \
    #    read_crystal_students_of_groups_pack('mipt_very')
    #print('+ mipt_very: ' + str(len(list(set(mipt_crystal_1_gusm)))))
    #mipt_crystal_1_gusm += \
    #    read_crystal_students_of_groups_pack('mipt_roughly')
    #print('+ mipt_roughly: ' + str(len(list(set(mipt_crystal_1_gusm)))))
    #mipt_crystal_1_gusm = list(set(mipt_crystal_1_gusm))

    # Search crystal phystechs among friends of crystal
    # WARNING: execution of this code block lasts > 2 hour
    #new_crystal = load_crystal_students_from_friends(mipt_crystal_1_gusm,
    #                                             mipt_crystal_1_gusm,
    #                                             univer_ids)
    #mipt_crystal_1_gusm += new_crystal
    #print('+ mipt_friends: ' + str(len(list(set(mipt_crystal_1_gusm)))))
    #while len(new_crystal) > 0:
    #    new_crystal = \
    #        load_crystal_students_from_friends(new_crystal,
    #                                       mipt_crystal_1_gusm,
    #                                       univer_ids)
    #    mipt_crystal_1_gusm += new_crystal
    #    print('+ mipt_friends: ' + str(len(list(set(mipt_crystal_1_gusm)))))

    # Write students to data/users_packs/
    #vk.write_users_pack(mipt_crystal_1_gusm, 'mipt_crystal_1_gusm')

    # Search students among friends of crystal on the basis of
    # the number of crystals in friends
    #mipt_crystal_1_gusm = vk.read_users_pack('mipt_crystal_1_gusm')
    #cryst_friends_cnt = load_friends_of_users_with_cnt(mipt_crystal_1_gusm)
    #print('cryst_friends_cnt: ' + str(len(cryst_friends_cnt)))
    #cryst_friends_2 = \
    #    [x for x in cryst_friends_cnt if cryst_friends_cnt[x] >= 2]
    #vk.write_users_pack(cryst_friends_2, 'cryst_friends_2')
    #print('cryst_friends_2: ' + str(len(cryst_friends_2)))
    #cryst_friends_3 = \
    #    [x for x in cryst_friends_cnt if cryst_friends_cnt[x] >= 3]
    #vk.write_users_pack(cryst_friends_3, 'cryst_friends_3')
    #print('cryst_friends_3: ' + str(len(cryst_friends_3)))
    #cryst_friends_5 = \
    #    [x for x in cryst_friends_cnt if cryst_friends_cnt[x] >= 5]
    #vk.write_users_pack(cryst_friends_5, 'cryst_friends_5')
    #print('cryst_friends_5: ' + str(len(cryst_friends_5)))
    #cryst_friends_10 = \
    #    [x for x in cryst_friends_cnt if cryst_friends_cnt[x] >= 10]
    #vk.write_users_pack(cryst_friends_10, 'cryst_friends_10')
    #print('cryst_friends_10: ' + str(len(cryst_friends_10)))

    #vk.load_users_pack_by_groups_pack('thematic_very')
    #vk.load_users_pack_by_groups_pack('thematic_roughly')
    #vk.load_users_pack_by_groups_pack('thematic_side') # TODO Докачать

    # Read phystechs and thematic from dump
    mipt_crystal = set(vk.read_users_pack('mipt_crystal_1_gusm'))
    cryst_friends_10 = set(vk.read_users_pack('cryst_friends_10'))
    cryst_friends_5 = set(vk.read_users_pack('cryst_friends_5'))
    thematic_very = set(vk.read_users_pack('thematic_very'))
    exceptions = set(vk.read_users_pack('exceptions'))

    intersection = list((mipt_crystal & thematic_very) - exceptions)
    intersection.sort()
    extended_intersection = \
        make_users_list_with_add_info(intersection,
                                      ['mipt_crystal_gr', 'mipt_very'],
                                      ['thematic_very', 'thematic_roughly'])
    make_thematic_report(extended_intersection, univer_ids)

    #intersection = list((cryst_friends_10 & thematic_very) - exceptions)
    #intersection.sort()
    #extended_intersection = \
    #    make_users_list_with_add_info(intersection,
    #                             ['mipt_crystal_gr', 'mipt_very'],
    #                             ['thematic_very', 'thematic_roughly'])
    #make_thematic_report(extended_intersection, univer_ids)
    #sys.exit()

    #intersection = list(set(cryst_friends_5) & set(thematic_very))
    #intersection.sort()
    #print(len(intersection))
    #print(intersection)



#! /usr/bin/env python3

"""Module for creating and sending program performance report
in .txt and .html formats"""

from .usrdata import UsrData
import time
import os
import sys
import smtplib
from email.mime.text import MIMEText
from email.header import Header
# TODO: Сохранять текст сообщений email, которые не удалось отправить


class Report():
    def __init__(self, report_type, html_only=True):
        self.type = report_type
        self.short_name = time.strftime("%F-%H%M%S")
        self.name = 'reports/' + self.type + '/' + self.short_name + '.html'
        self.text = 'Report is saved in ' + self.name + '\n\n'
        self.html = '\n'.join(
            '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">',
            '<Html>',
            '<Head>',
            '<Title>' + self.name + '</Title>',
            '<meta http-equiv="Content-Type" content="text/html;'
            + ' charset=UTF-8">',
            '</Head>',
            '<Body topmargin="0" leftmargin="0" rightmargin="0"'
            + ' bottommargin="0" marginheight="0" marginwidth="0">') + '\n'
        self.html_only = html_only
        self.empty = True

    def add_line(self, line):
        if not self.html_only:
            self.text += line + '\n'
        self.html += line + '<br>' + '\n'
        self.empty = False

    def add_str(self, s):
        if not self.html_only:
            self.text += s
        self.html += s
        self.empty = False

    def add_vk_link(self, addr, name):
        if not self.html_only:
            self.text += '[https://vk.com/' + addr + '] ' + name + '\n'
        self.html += f'<a href=https://vk.com/{addr}>{name}</a><br>\n'
        self.empty = False

    def conclude(self):
        self.html += '</Body>\n</Html>\n'

    def broadcast(self):

        # Get account data
        _, ac_obj = UsrData().get_active_obj('acc', 'email')
        sender, pswd = ac_obj['uname'], ac_obj['password']

        # Read emails of recipients
        recipients = UsrData().get('adm', 'bc_emails')
        if not recipients:
            print('No monitoring groups (try command monitor_add)')
            sys.exit()

        # Create message
        msg = MIMEText(self.text, 'plain', 'utf-8')
        msg['Subject'] = Header('Report ' + self.short_name, 'utf-8')
        msg['From'] = sender
        msg['To'] = ', '.join(recipients)

        s = smtplib.SMTP('smtp.gmail.com', 587, timeout=10)
        #s.set_debuglevel(1)
        try:
            s.starttls()
            s.login(sender, pswd)
            s.sendmail(msg['From'], recipients, msg.as_string())
        finally:
            #print(msg)
            s.quit()

    def dump(self):
        print(self.text)
        if not os.path.isdir('reports/'):
            os.mkdir('reports/')
        if not os.path.isdir('reports/' + self.type):
            os.mkdir('reports/' + self.type)
        with open(self.name, 'w') as f:
            f.write(self.html)

    def is_empty(self):
        return self.empty



#! /usr/bin/env python3

"""Realisation of API requests to vk.com. For single request use
`apply_vk_method`. For speedup you can use class `Executor`, wich pack
many requests to packs by 25 requests."""

import requests
import json
import time
import os
import getpass
from . import vkauth
from ..usrdata import UsrData
import logging
from logging.handlers import RotatingFileHandler

# Adjust logging
handler = RotatingFileHandler(".vklib.vkreq.log", mode='a',
                              maxBytes=1024*1024*10, backupCount=1)
handler.setFormatter(logging.Formatter("%(asctime)s vkreq: %(message)s"))
handler.setLevel(logging.DEBUG)
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.addHandler(handler)

_token = None
mock_responses = '.mock_request_responses.json'


def update_token():
    """Read token from local user data or get new one from vk.com"""

    global _token
    u = UsrData()

    # For first call
    if not _token:
        # Try read token from the previous run
        ac_name, ac_obj = u.get_active_obj('acc', 'vk')
        _token = ac_obj['token']
        if _token:
            logger.debug("Token is read")
            return

    # Get token by request
    _token = None
    attempts = 5
    attempt_pause = 10
    while not _token and attempts > 0:

        # Wait for sparse requests
        if attempts < 5:
            logger.debug("Attempt pause %s", attempt_pause)
            time.sleep(attempt_pause)
            attempt_pause *= 2
        attempts -= 1

        # Request
        logger.debug("Try to get token")
        ac_name, ac_obj = u.get_active_obj('acc', 'vk')
        email = ac_obj['uname']
        pswd = ac_obj['password']
        while not pswd:
            pswd = getpass.getpass('Input password of vk account'
                                   + ' «{}»: '.format(ac_name))
        user_auth = vkauth.VKAuth(permissions=['friends',
                                               'groups',
                                               'wall'],
                                  app_id='6471192',
                                  api_v='5.74',
                                  email=email,
                                  pswd=pswd)
        user_auth.auth()
        _token = user_auth._access_token

    # Here token must be
    if _token is None:
        logger.debug("Failed to get token")
        raise

    # Dump token (if account isn't private)
    if ac_obj['password']:
        u.set(_token, 'acc', 'vk', ac_name, 'token')
    logger.debug("Token is updated")


def _set_token_to_params(params):

    # Add token & version (if user doesn't give specific this)
    if 'access_token' not in params:
        params['access_token'] = _token
    if 'v' not in params:
        params['v'] = '5.9'


def _make_params_string(params):
    """Convert list of parameters to string"""

    p_str = '&'.join(['{}={}'.format(f, params[f]) for f in params.keys()])
    return p_str


def _short_print(s):
    """shrink output if it's too long"""
    if len(s) <= 500:
        print(s)
    else:
        print('{}  ......  {}'.format(s[:245], s[-245:]))


def _vk_api_error_print(obj, url_of_req, params):
    print("vk API error code: {}".format(obj['error_code']))
    print(obj['error_msg'])
    _short_print("Broken request: {}{}".format(url_of_req,
                                               _make_params_string(params)))


def _vk_api_request(url_of_req, method, params):
    """Request (can be mocked by existence of file
    '.mock_request_responses.json' for testing)."""

    if os.path.isfile(mock_responses):

        # "Home-made" mock
        logger.debug("Try Mocked Request (method: %s)", method)
        with open(mock_responses) as fp:
            mocks_list = json.load(fp)
            json_obj = mocks_list.pop(0)
        with open(mock_responses, 'w') as fp:
            json.dump(mocks_list, fp)
        return json_obj

    else:

        # Real request
        logger.debug("Try Request (method: %s)", method)
        response = requests.post(url_of_req, data=params)
        return response.json()


def apply_vk_method(method, handle_api_errors=True, **params):
    """Make request to https://api.vk.com/method/. Return JSON-object"""

    global _token

    # Get token if absent
    if not _token and not os.path.isfile('.mock_request_responses.json'):
        update_token()

    # Set token to URL string
    url_of_req = 'https://api.vk.com/method/' + method + '?'
    _set_token_to_params(params)

    # Do request with error processing
    error_pause = 5
    while True:

        try:
            # Request
            json_obj = _vk_api_request(url_of_req, method, params)
        except Exception as e:
            # In case of network problems.
            logger.debug("Request Error: " + str(e))
            p_str = _make_params_string(params)
            _short_print("Broken request: {}{}".format(url_of_req, p_str))
            print(e)
            if error_pause > 35:
                raise
            print('Wait ' + str(error_pause) + ' seconds')
            time.sleep(error_pause)
            error_pause += 5
            continue

        # Process response
        if 'error' in json_obj:
            logger.debug("API Error (%s: %s)", json_obj['error']['error_code'],
                         json_obj['error']['error_msg'])
            if json_obj['error']['error_code'] == 5:
                if 'access_token' in params:
                    del params['access_token']
                update_token()
                _set_token_to_params(params)
                continue
            elif json_obj['error']['error_code'] == 6:
                # It took a long time to create a "clever" mechanism to bypass
                # these errors. However, such a stupid solution was the most
                # effective.
                time.sleep(0.5)
                continue
            elif not handle_api_errors:
                return json_obj
            elif json_obj['error']['error_code'] == 18:
                # Page is deleted or banned
                return {}
            elif (json_obj['error']['error_code'] == 203
                  or json_obj['error']['error_code'] == 7):
                # No have access
                return {}
            elif json_obj['error']['error_code'] == 12:
                # execute compilation error
                _vk_api_error_print(json_obj['error'], url_of_req, params)
                return {}
            elif json_obj['error']['error_code'] == 14:
                # Captcha needed
                _vk_api_error_print(json_obj['error'], url_of_req, params)
                return json_obj  # to study the response format in the future
            elif json_obj['error']['error_code'] in (3, 8, 100, 113):
                # Wrong rquest
                _vk_api_error_print(json_obj['error'], url_of_req, params)
                return {}
            else:
                # In case of server problems.
                _vk_api_error_print(json_obj['error'], url_of_req, params)
                if error_pause > 35:
                    raise
                print('Wait ' + str(error_pause) + ' seconds')
                time.sleep(error_pause)
                error_pause += 5
                continue

        # No errors -> out loop
        logger.debug("Successful Request")
        break

    return json_obj


class Executor:
    """Class for exploit vk API method `execute` to join
    several requests (up to 25) into single one.
    If > 25 requests are added by `add_request`, then > 1 requests
    of `execute` will be executed by `emit_requests`.

    Example:
    >>> e = Executor()
    >>> e.add_request('utils.resolveScreenName', screen_name='phys_kek')
    >>> e.add_request('utils.resolveScreenName_', screen_name='phys_kek')
    >>> e.add_request('users.getFollowers', **{'user_id': '1'})
    >>> e.emit_requests()
    >>> e.responses[0]['object_id'], e.responses[1], e.responses[2]['count']
    (111557607, [], 5909523)

    More tricky way - use parameter `processor` of `Executor.add_request()`:
    >>> e = Executor()
    >>> f = lambda r: print(r['object_id'] if r else "False")
    >>> for name in ['phys_kek', 'unreal-_-', 'drec_mipt']:
    >>>     e.add_request('utils.resolveScreenName', f, screen_name=name)
    >>> e.emit_requests()
    111557607
    False
    17708"""

    def __init__(self):
        logger.debug("Executor created")
        self.responses = []
        self.errors = []
        self.requests = []
        self.processors = []

    def add_request(self, method, processor=None, **params):
        """Add one request to `Executor.requests`.
        `processor` - function for processing response of this request"""

        logger.debug("Add %sth request %s", len(self.requests) + 1,
                     "with processor" if processor else "without processor")

        # add string request to requests list
        params_s = str(params).replace("'", '"')
        self.requests.append('API.{}({});'.format(method, params_s))
        self.processors.append(processor)

    def emit_requests(self):
        """Pack requests from `Executor.requests`
        (see `Executor.add_request()`) into `execute` requests and emit them
        to server. Responses will be saved in `Executor.responses`."""

        logger.debug("Emit requests (%s pieces)", len(self.requests))
        self.responses = []
        self.errors = []

        requests = self.requests
        cnt = 0
        while requests:

            # start code for execute
            if cnt == 0:
                code = 'var arr = [];\n' + \
                       'var r;\n'

            # collect requests
            req = requests.pop(0)
            code += 'r = {}\n'.format(req) + \
                    'arr = arr + [r];\n'

            # step - recorded regular request
            cnt += 1

            # check for maximum requests to execute
            if cnt == 25:

                # complete code
                code += 'return arr;'
                r = apply_vk_method('execute', **{'code': code})

                # save responses
                self.responses += r['response']
                if 'execute_errors' in r:
                    self.errors += r['execute_errors']

                # process responses
                while cnt:
                    func = self.processors.pop(0)
                    if func:
                        func(r['response'][-cnt])
                    cnt -= 1

        # make last execute if need
        if cnt:

            # complete code
            code += 'return arr;'
            r = apply_vk_method('execute', **{'code': code})

            # save responses
            self.responses += r['response']
            if 'execute_errors' in r:
                self.errors.append(r['execute_errors'])

            # process responses
            while cnt:
                func = self.processors.pop(0)
                if func:
                    func(r['response'][-cnt])
                cnt -= 1

        self.requests = []
        self.processors = []



#! /usr/bin/env python3

"""Module for working with data stored in the directory .vkts/.
These include user accounts, emails, communities and universities
of interest to user, topics, etc."""

import os
import json
from .utils import exception_handler


class UsrData:

    # paths to data files
    data_path = '.vkts'
    acc_path = os.sep.join((data_path, 'accounts.json'))
    adm_path = os.sep.join((data_path, 'adm_data.json'))
    univ_path = os.sep.join((data_path, 'univers.json'))
    # data type file name -> file path
    path_map = {'acc': acc_path,
                'adm': adm_path,
                'univ': univ_path}

    def generate_if_absent(self):
        """Generate user data blank if absent"""

        if not os.path.isdir(self.data_path):
            os.mkdir(self.data_path)
        if not os.path.isfile(self.acc_path):
            #acc_obj = {"email": None, "vk": None, "telegram": None}
            json.dump({}, open(self.acc_path, 'w'))
        if not os.path.isfile(self.adm_path):
            adm_obj = {"bc_emails": [], "mon_groups": []}
            json.dump(adm_obj, open(self.adm_path, 'w'))
        if not os.path.isfile(self.univ_path):
            json.dump({}, open(self.univ_path, 'w'))

    def open(self):
        """Open user data"""

        try:
            self.acc = json.load(open(self.acc_path))
            self.adm = json.load(open(self.adm_path))
            self.univ = json.load(open(self.univ_path))
        except Exception as e:
            exception_handler(e, 'Incorrect data in ' + self.data_path)

    def get(self, *field_path):
        """Safe get object of UsrData. For instance:

         >>> u = UsrData()
         >>> is_act = u.get('acc', 'vk', 'charm', 'is_activated')

        to know, is account `charm` activated"""

        self.open()

        try:
            # go down the data structure untill target (or None if absent)
            d = self.__dict__
            fields = list(field_path)
            for f in fields:
                if d is None:
                    return None
                if isinstance(f, int):
                    assert(isinstance(d, list))
                    if f < len(d):
                        d = d[f]
                    else:
                        return None
                else:
                    if f in d:
                        d = d[f]
                    else:
                        return None
            return d

        except Exception as e:
            exception_handler(e, 'Reading user data error')

    def set(self, new_obj, *field_path, correct_is_act=False):
        """Safe write to UsrData instance with data file update.
        For instance do:

          >>> u = UsrData()
          >>> u.set(True, 'acc', 'vk', 'charm', 'is_activated')

        to write `True` into u.acc['vk']['charm']['is_activated']"""

        self.generate_if_absent()
        self.open()

        try:
            # look for/create the specified place in the object
            d = self.__dict__
            fields = list(field_path)
            last = fields.pop()
            for f in fields:
                if f not in d or d[f] is None:
                    d[f] = {}
                d = d[f]

            # update user data in memory
            if last in d and isinstance(d[last], list):
                d[last].append(new_obj)
            else:
                d.update(((last, new_obj),))

            # Correct attribut `is_activated`
            # (activate random if there is a single object)
            if correct_is_act and len(d) > 0:
                attr_list = [x['is_activated'] for x in d.values()]
                if not list(filter(bool, attr_list)):
                    name = list(d.keys())[0]
                    d[name]['is_activated'] = True

            # update user data in data file
            data_type = field_path[0]
            file_path = self.path_map[data_type]
            json.dump(self.__dict__[data_type], open(file_path, 'w'))

        except Exception as e:
            exception_handler(e, 'Saving user data error')

    def del_(self, *field_path, correct_is_act=False):
        """Safe delete object from UsrData with data file update.
        For instance:

         >>> u = UsrData()
         >>> u.del_('acc', 'vk', 'charm')

        to delete `charm`"""

        self.generate_if_absent()
        self.open()

        try:
            # go down the data structure untill target (or None if absent)
            d = self.__dict__
            fields = list(field_path)
            last = fields.pop()
            for f in fields:
                if d is None:
                    return None
                if isinstance(f, int):
                    assert(isinstance(d, list))
                    if f < len(d):
                        d = d[f]
                    else:
                        return None
                else:
                    if f in d:
                        d = d[f]
                    else:
                        return None

            # del
            if isinstance(d, list):
                if isinstance(last, int):
                    if last < len(d):
                        #print('list[int]: {}, {}'.format(d, last))
                        del d[last]
                else:
                    if last in d:
                        #print('list.remove(str): {}, {}'.format(d, last))
                        d.remove(last)
            elif isinstance(d, dict):
                if last in d:
                    #print('dict[str]: {}, {}'.format(d, last))
                    del d[last]

            # Correct attribut `is_activated`
            # (activate random if there is a single object)
            if correct_is_act and len(d) > 0:
                attr_list = [x['is_activated'] for x in d.values()]
                if not list(filter(bool, attr_list)):
                    name = list(d.keys())[0]
                    d[name]['is_activated'] = True

            # update user data in data file
            data_type = field_path[0]
            file_path = self.path_map[data_type]
            json.dump(self.__dict__[data_type], open(file_path, 'w'))

        except Exception as e:
            exception_handler(e, 'Deleting user data error')

    def drop_activations(self, *field_path):
        """Set `False` to all objects in objects dictionary"""

        self.generate_if_absent()
        self.open()

        # get objects dictionary
        objs_dict = self.get(*field_path)

        try:
            # drop attribute
            for obj_key in objs_dict.keys():
                objs_dict[obj_key]['is_activated'] = False

            # update user data in data file
            data_type = field_path[0]
            file_path = self.path_map[data_type]
            json.dump(self.__dict__[data_type], open(file_path, 'w'))

        except Exception as e:
            exception_handler(e, 'Droping attribute error in user data')

    def get_active_obj(self, *field_path):
        """Find activated objects on end of field path.
        Returns: object_name, object"""

        self.open()

        # get objects dictionary
        objs_dict = self.get(*field_path)

        try:
            # find activated object
            for obj_key in objs_dict.keys():
                if objs_dict[obj_key]['is_activated']:
                    break
            else:
                raise

            return obj_key, objs_dict[obj_key]

        except Exception as e:
            exception_handler(e, '\n'.join(
                'Active {} account is not found.',
                '(Maybe need to execute command ac_add for',
                'adding account of type \'vk\' is needed or',
                'other *_add command)'))



#! /usr/bin/env python3

"""Module for administering and monitoring vk.com communities"""

import os
import sys
import time
from .report import Report
from . import vklib as vk
from .vklib import apply_vk_method
from .usrdata import UsrData


def add_group_for_monitoring(group_id):
    """Add group_id in registry of monitoring groups"""

    u = UsrData()

    # make group object fo saving
    digit_id, domain = vk.resolve_group_ids(group_id)
    obj = {'id': digit_id, 'domain': domain}

    # avoid repetition
    obj_list = u.get('adm', 'mon_groups')
    for o in obj_list:
        if o['id'] == digit_id:
            return

    # dump administrative data
    u.set(obj, 'adm', 'mon_groups')


def remove_monitoring_group(group_id):
    """Remove monitoring group_id"""

    # get list of objects
    u = UsrData()
    objs_list = u.get('adm', 'mon_groups')
    assert isinstance(objs_list, list), 'Monitoring groups data is incorrect'

    # look for object with specific id
    is_digit = (isinstance(group_id, int) or group_id.isdigit())
    for i, obj in enumerate(objs_list):
        assert isinstance(obj, dict), \
                'Monitoring groups data is incorrect: {}'.format(obj)
        # delete it from list
        if ((is_digit and obj['id'] == int(group_id))
                or obj['domain'] == group_id):
            u.del_('adm', 'mon_groups', i)
            break


def display_monitoring_groups():
    """Print all monitoring group_id"""

    objs_list = UsrData().get('adm', 'mon_groups')
    if objs_list:
        for obj in objs_list:
            print('{} \t{}'.format(obj['id'], obj['domain']))
        print('')


def add_email_for_broadcasting(email):
    """Add email address for broadcast reports about monitoring groups"""

    u = UsrData()

    # avoid repetition
    emails_list = u.get('adm', 'bc_emails')
    if emails_list and email in emails_list:
        return

    # add email
    u.set(email, 'adm', 'bc_emails')


def remove_broadcasting_email(email):
    """Remove email from addresses for broadcast reports
    about monitoring groups"""

    UsrData().del_('adm', 'bc_emails', email)


def display_broadcasting_email():
    """Print all addresses for broadcast reports about monitoring groups"""

    adm_data = UsrData().get('adm')
    if 'bc_emails' in adm_data and adm_data['bc_emails']:
        print('{}\n'.format('\n'.join(adm_data['bc_emails'])))


def check_updates(broadcast):
    """Check for changes in target groups, make report, send it to email"""

    # TODO: _mkdir_rec
    # TODO: Уменьшить число запросов
    tg_dir = os.sep.join(['data', 'adm_groups']) + os.sep

    # Read list of groups ids
    gr_list = UsrData().get('adm', 'mon_groups')
    if not gr_list:
        print('No monitoring groups (try command monitor_add)')
        sys.exit()
    gr_list = [[x['id'], x['domain']] for x in gr_list]

    # Make report
    r = Report('adm_groups_updates', html_only=False)
    new_files = []
    for g_info in gr_list:

        # Check accessibility of group data
        res = apply_vk_method('groups.getMembers', group_id=g_info[0])
        if 'error' in res and res['error']['error_code'] == 203:
            continue

        # Get data about group
        g = vk.Group(g_info[0])
        try:
            g.load()
        except Exception as e:
            # In case of network problems we must delete collected data.
            # Otherwise, this data will spoil the next attempt
            # to generate a report
            print('Could not load information about the group of ' + g_info[1]
                  + ' (' + str(e) + ')')
            for f in new_files:
                os.remove(f)
            sys.exit()
        if g.is_empty():
            continue

        # Dump data
        if not os.path.isdir(tg_dir + g_info[1] + '/'):
            os.mkdir(tg_dir + g_info[1] + '/')
        if not os.path.isdir(tg_dir + g_info[1] + '/members/'):
            os.mkdir(tg_dir + g_info[1] + '/members/')
        out_file_name = (tg_dir + g_info[1] + '/members/'
                         + time.strftime("%F-%H%M%S") + '.txt')
        with open(out_file_name, 'w') as f:
            for user in g.members:
                f.write(str(user) + '\n')

        # Diff 2 last files
        ls_res = os.listdir(tg_dir + g_info[1] + '/members/')
        if len(ls_res) < 2:
            continue
        ls_res.sort()
        old_file = tg_dir + g_info[1] + '/members/' + ls_res[-2]
        new_file = tg_dir + g_info[1] + '/members/' + ls_res[-1]
        with open(old_file, 'r') as f:
            old_list = f.read().split('\n')[:-1]
        with open(new_file, 'r') as f:
            new_list = f.read().split('\n')[:-1]
        just_in_old = list(set(old_list) - set(new_list))
        just_in_new = list(set(new_list) - set(old_list))

        # If last 2 filaes are the same, then delete last one
        if len(just_in_old) == 0 and len(just_in_new) == 0:
            os.remove(new_file)
            continue
        else:
            new_files.append(out_file_name)
            if just_in_old:
                r.add_line('Покинули группу ' + g_info[1] + ':')
                for user_id in just_in_old:
                    r.add_str('[' + user_id + '] ')
                    r.add_vk_link('id' + user_id,
                                  vk.get_user_name(user_id))
                r.add_line('')
            if just_in_new:
                r.add_line('Вошли в группу ' + g_info[1] + ':')
                for user_id in just_in_new:
                    r.add_str('[' + user_id + '] ')
                    r.add_vk_link('id' + user_id,
                                  vk.get_user_name(user_id))
                r.add_line('')
            if just_in_old or just_in_new:
                r.add_line('В группе ' + str(g.count) + ' человек')
                r.add_line('')

    # Close report and broadcast it
    if not r.is_empty():
        r.conclude()
        r.dump()
        if broadcast:
            r.broadcast()


# Make report about birthday of all members of groups
def make_birthday_calendar(group_id):

    # Read arg
    group_id, scrname = vk.resolve_group_ids(group_id)

    # Load info about members
    g = vk.Group(group_id)
    g.load()
    members = g.members
    btd_list = []
    step = 200  # instead of 1000 to avoid error 414 (URI Too Long)
    last_index = int((len(members)-1)/step)
    for i in range(0, last_index + 1):
        if i == last_index:  # TODO: Why `count` is unused?
            count = len(members) % step
        else:
            count = step
        members_str = ','.join([str(user)
                                for user in members[i*step:i*step+step]])
        btd_list += apply_vk_method('users.get', user_ids=members_str,
                                    fields='bdate,photo_max')['response']

    # Empty list -> out
    if not btd_list:
        print('No members with known date of birthday')
        sys.exit()

    # Delete users without birthday date
    tmp = [user for user in btd_list if 'bdate' in user]
    btd_list = tmp

    # Set default ava if it's absent
    for user in btd_list:
        if not user['photo_max']:
            user['photo_max'] = 'https://vk.com/images/camera_200.png?ava=1'

    # Add fields 'bmonth' and 'bday'
    for user in btd_list:
        user.update({'bmonth': int(user['bdate'].split('.')[1]),
                     'bday': int(user['bdate'].split('.')[0])})

    # Add field 'bval_for_sort' = 'bmonth' * 100 + 'bdate'
    for user in btd_list:
        user.update({'bval_for_sort': user['bmonth'] * 100 + user['bday']})

    # Sort by birthday date
    btd_list.sort(key=(lambda user: user['bval_for_sort']))

    # Create calendar report
    r = Report('birthday_calendar')
    i = 0
    r.add_str('<table>\n')
    for user in btd_list:
        i += 1
        r.add_str('<tr>\n')
        r.add_str(' <td valign="top">' + str(i) + '.</td>\n')
        r.add_str(' <td><a href="https://vk.com/id' + str(user['id'])
                  + '" target="_blank"><img height=50 src="'
                  + user['photo_max'] + '"></a></td>\n')
        r.add_str(' <td valign="top">\n')
        r.add_line('  ' + user['first_name'] + ' ' + user['last_name'])
        months = ['января', 'февраля', 'марта', 'апреля',
                  'мая', 'июня', 'июля', 'августа',
                  'сентября', 'октября', 'ноября', 'декабря']
        r.add_line('  ' + str(user['bday']) + ' ' + months[user['bmonth']-1])
        r.add_str('</td>\n')
        r.add_str('</tr>\n')
    r.add_str('</table>\n')

    # Conclude report
    if not r.is_empty():
        r.conclude()
        r.dump()



#! /usr/bin/env python3

"""Most frequently used requests to vk.com"""

from .vkreq import apply_vk_method, Executor
from collections.abc import Iterable


def get_group_domain(group_id):
    response = apply_vk_method('groups.getById', group_id=group_id)
    return response['response'][0]['screen_name']


def get_group_name(group_id):
    response = apply_vk_method('groups.getById', group_id=group_id)
    return response['response'][0]['name']


def resolve_group_ids(group_ids):
    """resolve_group_ids(digit id or domain of group) -> digit_id, domain_id.
    You can pass iterable argument for resolve every id in it."""

    if isinstance(group_ids, Iterable) and not isinstance(group_ids, str):

        # load info for resolving
        e = Executor()
        for group_id in group_ids:
            # is `group_id` digit or domain?
            if isinstance(group_id, int) or group_id.isdigit():
                e.add_request('groups.getById', group_ids=group_id)
            else:
                e.add_request('utils.resolveScreenName', screen_name=group_id)
        e.emit_requests()

        # collect result list & return
        res = []
        for group_id, r in zip(group_ids, e.responses):
            # is `group_id` digit or domain?
            if isinstance(group_id, int) or group_id.isdigit():
                digit_id = int(group_id)
                domain_id = r[0]['screen_name']
            else:
                digit_id = r['object_id']
                domain_id = group_id
            res.append((digit_id, domain_id))
        return res

    # `group_ids` is digit or domain?
    if isinstance(group_ids, int) or group_ids.isdigit():
        digit_id = int(group_ids)
        response = apply_vk_method('groups.getById', group_ids=group_ids)
        domain_id = response['response'][0]['screen_name']
    else:
        response = apply_vk_method('utils.resolveScreenName',
                                   screen_name=group_ids)
        digit_id = response['response']['object_id']
        domain_id = group_ids

    return digit_id, domain_id


def get_user_name(user_id):
    response = apply_vk_method('users.get', user_ids=user_id)
    if response:
        return response['response'][0]['first_name'] + ' ' \
               + response['response'][0]['last_name']
    else:
        return 'No name'


def resolve_user_ids(user_id):
    """resolve_user_ids(digit id or domain of user) -> digit_id, domain_id"""

    # `user_id` is digit or domain?
    if isinstance(user_id, int) or user_id.isdigit():
        digit_id = int(user_id)
        response = apply_vk_method('users.get', user_ids=user_id,
                                   fields='screen_name')
        domain_id = response['response'][0]['screen_name']
    else:
        response = apply_vk_method('utils.resolveScreenName',
                                   screen_name=user_id)
        digit_id = response['response']['object_id']
        domain_id = user_id

    return digit_id, domain_id


def domain_2_digital_id(domain):
    response = apply_vk_method('utils.resolveScreenName', screen_name=domain)
    return response['response']['object_id']



#! /usr/bin/env python3

"""Realisation of access to the functionality of `vkts` via the console"""

import sys
from . import real
from . import admreal as adm
from . import vklib as vk

commands_with_descr = \
  [
    [
      'Account commands',

      ['ac_add', '   # enter interactive mode'],
      ['ac_rem', '[email/vk/telegram] <ac_name>'],
      ['ac_activate', '[email/vk/telegram] <ac_name>'],
      ['ac_see', '']
    ],
    [
      'University commands',

      ['un_add', '   # enter interactive mode'],
      ['un_rem', '<un_name>'],
      ['un_activate', '<un_name>'],
      ['un_see', '']
    ],
    [
      'VK API commands',

      ['method', '<method_name> {param0=<v0> param1=...}'],
      ['resolve_group_ids', '<screen_name>'],
      ['resolve_user_ids', '<group_id>']
    ],
    [
      'Commands for working with packs',

      ['grouppack_add', '<pack_name> <group_id>'],
      ['grouppack_rem', '<pack_name> <group_id>'],
      ['exceptpack_add', '<user_id> {comment}'],
      ['exceptpack_rem', '<user_id>'],
      ['search_phystech_groups', '']  # TODO:
      # Искать по ключ. словам и/или спискам групп физтехов группы,
      # вычислять их mipt_degree и вывоить в сортированном порядке
      # (добавить опцию с диапазоном degree для включения в заданный
      #  другой опцией groups_pack; а лучше сделать одной опцией для
      #  пополнения одной командой сразу нескольких pack-ов)
    ],
    [
      'Elementary analytical commands',

      ['load_users_pack_by_groups_pack', ''],
      ['load_phystechs_by_search', ''],
      ['drop_groups_data_relevance', ''],
      ['drop_users_data_relevance', '']
    ],
    [
      'Complete analysis',

      ['get_thematic_phystechs_by_their_groups', ''],
      ['get_thematic_phystechs_by_their_wall', '']  # TODO
    ],
    [
      'Administrative commands',

      ['monitor_add', '<group_id>'],
      ['monitor_rem', '<group_id>'],
      ['monitor_see', ''],
      ['broadcast_add', '<email>'],
      ['broadcast_rem', '<email>'],
      ['broadcast_see', ''],
      ['check_updates', '{--no-broadcast}'],
      ['make_birthday_calendar', '<group_id>']
    ]
  ]

# Make list of command names
commands = [y[0] for x in commands_with_descr for y in x[1:]]


def main():

    # Print Help
    if (len(sys.argv) == 1
            or sys.argv[1] not in commands):
        for chapter in commands_with_descr:
            print('')
            print(chapter[0])
            for item in chapter[1:]:
                print(' '.join(['vkts'] + item))
        print('')
        sys.exit()

    # Execute command
    if sys.argv[1] == 'ac_add':
        real.add_account()
    elif sys.argv[1] == 'ac_rem':
        real.delete_account(sys.argv[2], sys.argv[3])
    elif sys.argv[1] == 'ac_activate':
        real.activate_account(sys.argv[2], sys.argv[3])
    elif sys.argv[1] == 'ac_see':
        real.display_accounts()

    elif sys.argv[1] == 'un_add':
        real.add_university()
    elif sys.argv[1] == 'un_rem':
        real.delete_university(sys.argv[2])
    elif sys.argv[1] == 'un_activate':
        real.activate_university(sys.argv[2])
    elif sys.argv[1] == 'un_see':
        real.display_universities()

    elif sys.argv[1] == 'method':
        # Apply some method and output result
        print(real.vk_method(sys.argv[2], sys.argv[3:]))
    elif sys.argv[1] == 'resolve_group_ids':
        print(vk.resolve_group_ids(sys.argv[2]))
    elif sys.argv[1] == 'resolve_user_ids':
        print(vk.resolve_user_ids(sys.argv[2]))

    # TODO: Не должно быть отдельных опций для исключений
    # TODO: Имена опций неудачные
    # TODO: опций, специфичных для конр. ВУЗа, не должно быть
    elif sys.argv[1] == 'grouppack_add':
        # Add group in data/groups_packs/pack by id or scrname
        vk.add_group_to_pack(sys.argv[2], sys.argv[3])
    elif sys.argv[1] == 'grouppack_rem':
        # Remove group from data/groups_packs/pack by id or scrname
        vk.rem_group_from_pack(sys.argv[2], sys.argv[3])
    elif sys.argv[1] == 'exceptpack_add_user':
        # Add group in data/groups_packs/pack by id or scrname
        vk.add_user_to_pack(sys.argv[2], 'exceptions', sys.argv[3])
    elif sys.argv[1] == 'exceptpack_rem_user':
        # Remove group from data/groups_packs/pack by id or scrname
        vk.rem_user_from_pack(sys.argv[2], 'exceptions')
    elif sys.argv[1] == 'search_phystech_groups':
        print('This command is not realised')

    # TODO: Этот блок удалить после переделки работы с pack-ми
    elif sys.argv[1] == 'load_users_pack_by_groups_pack':
        # For groups in set data/groups_packs/pack_name load members list
        # and write it to data/users_packs/pack_name
        if (len(sys.argv) > 3):
            vk.load_users_pack_by_groups_pack(sys.argv[2],
                                              sys.argv[3].split(','))
        else:
            vk.load_users_pack_by_groups_pack(sys.argv[2])
    elif sys.argv[1] == 'load_phystechs_by_search':
        # Load list of phystechs using method users.search
        real.load_phystechs_by_search(sys.argv[2])
    elif sys.argv[1] == 'drop_groups_data_relevance':
        # Mark all data stored in data/groups/ irrelevant
        vk.drop_relevance_flag_of_all_groups()
    elif sys.argv[1] == 'drop_groups_data_relevance':
        # Mark all data stored in data/users/ irrelevant
        vk.drop_relevance_flag_of_all_users()

    elif sys.argv[1] == 'get_thematic_phystechs_by_their_groups':
        # Analyse by groups
        real.get_thematic_students_by_their_groups()

    elif sys.argv[1] == 'monitor_add':
        # Add group_id in registry of monitoring groups
        adm.add_group_for_monitoring(sys.argv[2])
    elif sys.argv[1] == 'monitor_rem':
        # Remove monitoring group_id
        adm.remove_monitoring_group(sys.argv[2])
    elif sys.argv[1] == 'monitor_see':
        # Print all monitoring group_id
        adm.display_monitoring_groups()
    elif sys.argv[1] == 'broadcast_add':
        # Add email address for broadcast reports about monitoring groups
        adm.add_email_for_broadcasting(sys.argv[2])
    elif sys.argv[1] == 'broadcast_rem':
        # Remove email from addresses for broadcast reports
        adm.remove_broadcasting_email(sys.argv[2])
    elif sys.argv[1] == 'broadcast_see':
        # Print all addresses for broadcast reports about monitoring groups
        adm.display_broadcasting_email()
    elif sys.argv[1] == 'check_updates':
        # Find changes in monitoring groups
        adm.check_updates('--no-broadcast' not in sys.argv)
    elif sys.argv[1] == 'make_birthday_calendar':
        # Make report with sorted list of birthday of community
        adm.make_birthday_calendar(sys.argv[2])

    else:
        print('Command ' + sys.argv[1] + ' is not supported')
        sys.exit()


if __name__ == '__main__':
    main()



#! /usr/bin/env python3

"""Interfaces for working with packages of groups and users"""

import sys
import os
from .vkobjs import User, Group
from .hotreqs import domain_2_digital_id, get_group_name, get_group_domain


# Add group in data/groups_packs/pack_name by group_id or screen_name
def add_group_to_pack(pack_name, some_group_id):

    # Read args
    if some_group_id.isdigit():
        group_id = some_group_id
        scrname = get_group_domain(group_id)
    else:
        scrname = some_group_id
        group_id = str(domain_2_digital_id(scrname))

    # Build dir if absent
    if not os.path.isdir('data/groups_packs'):
        os.mkdir('data/groups_packs')

    # Check existance of group
    ls_res = os.listdir('data/groups_packs/')
    for f_name in ls_res:
        if os.path.isfile('data/groups_packs/' + f_name):
            with open('data/groups_packs/' + f_name, 'r') as f:
                res = f.read()
            if ' # ' + scrname + ' # ' in res:
                print('Set ' + 'data/groups_packs/' + f_name
                      + ' already consist this group')
                sys.exit()

    # Add group in file
    new_line = (group_id + ' # ' + scrname + ' # '
                + get_group_name(group_id) + '\n')
    with open('data/groups_packs/' + pack_name, 'a') as f:
        f.write(new_line)


# Remove group from data/groups_packs/pack_name by group_id or screen_name
def rem_group_from_pack(pack_name, some_group_id):

    # Check existence of pack
    if not os.path.isfile('data/groups_packs/' + pack_name):
        print('data/groups_packs/' + pack_name + ' is absent')
        sys.exit()

    # Read pack
    with open('data/groups_packs/' + pack_name, 'r') as f:
        lines = f.read().split('\n')[:-1]

    if some_group_id.isdigit():
        # Parse numeric ids
        groups_ids = [x.split(' # ')[0] for x in lines]
    else:
        # Parse domains
        groups_ids = [x.split(' # ')[1] for x in lines]

    # Check existence of group in pack
    if some_group_id not in groups_ids:
        print('Group ' + str(some_group_id) +
              ' is not found in pack ' + pack_name)
        sys.exit()

    # Remove group
    index = groups_ids.index(some_group_id)
    lines.pop(index)

    # Rewrite pack
    with open('data/groups_packs/' + pack_name, 'w') as f:
        f.write('\n'.join(lines) + ('\n' if lines else ''))


# Return list of group_ids of given pack
def read_groups_pack(pack_name):
    if not os.path.isfile('data/groups_packs/' + pack_name):
        print('data/groups_packs/' + pack_name + ' is absent')
        sys.exit()
    with open('data/groups_packs/' + pack_name, 'r') as f:
        lines = f.read().split('\n')[:-1]
    groups = [x.split(' # ')[0] for x in lines]
    return groups


###   Interfaces for working with users packages   ###


# Add user in data/users_packs/pack_name by user_id or screen_name
def add_user_to_pack(some_user_id, pack_name, comment=''):

    # Read args
    if some_user_id.isdigit():
        user_id = some_user_id
    else:
        user_id = str(domain_2_digital_id(some_user_id))

    # Build dir if absent
    if not os.path.isdir('data/users_packs'):
        os.mkdir('data/users_packs')

    # Add user in pack
    new_line = user_id
    if comment:
        new_line += ' # ' + comment
    with open('data/users_packs/' + pack_name, 'a') as f:
        f.write(new_line)


# Remove user from data/users_packs/pack_name by user_id or screen_name
def rem_user_from_pack(some_user_id, pack_name):

    # Read args
    if some_user_id.isdigit():
        user_id = some_user_id
    else:
        user_id = str(domain_2_digital_id(some_user_id))

    # Check existence of pack
    if not os.path.isfile('data/users_packs/' + pack_name):
        print('data/users_packs/' + pack_name + ' is absent')
        sys.exit()

    # Read pack
    with open('data/users_packs/' + pack_name, 'r') as f:
        lines = f.read().split('\n')[:-1]

    # Parse numeric ids
    users_ids = [x.split(' # ')[0] for x in lines]

    # Check existence of user in pack
    if user_id not in users_ids:
        print('User ' + user_id + ' is not found in pack ' + pack_name)
        sys.exit()

    # Remove user
    index = users_ids.index(user_id)
    lines.pop(index)

    # Rewrite pack
    with open('data/users_packs/' + pack_name, 'w') as f:
        f.write('\n'.join(lines) + ('\n' if lines else ''))


# Write list of user ids (users_list) to file in data/users_packs/
def write_users_pack(users_list, pack_name):
    if not os.path.isdir('data/users_packs/'):
        os.mkdir('data/users_packs/')
    with open('data/users_packs/' + pack_name, 'w') as f:
        for user in users_list:
            f.write(str(user) + '\n')


# Get members list of users pack
def read_users_pack(pack_name):
    if not os.path.isfile('data/users_packs/' + pack_name):
        print('data/users_packs/' + pack_name + ' is absent')
        sys.exit()
    with open('data/users_packs/' + pack_name, 'r') as f:
        lines = f.read().split('\n')[:-1]
        users_list = [int(x.split(' # ')[0]) for x in lines]
    return users_list


###   Common interfaces for working with packages   ###


# For groups in set data/groups_packs/pack_name load members list
# and write it to data/users_packs/pack_name
def load_users_pack_by_groups_pack(pack_name, univer_ids=[]):

    # Get list of groups
    groups = read_groups_pack(pack_name)

    # Getting list of members of groups
    members_list = []
    for g_id in groups:
        g = Group(g_id, univer_ids)
        g.open()
        if g.is_empty():
            g.load()
            g.dump()
        members_list += g.members()

    # Delete repeats and sort
    members_list = list(set(members_list))
    members_list.sort()

    # Write members to data/users_packs/
    write_users_pack(members_list, pack_name)


# Write 'False' to field 'is_relevant' of all stored groups
def drop_relevance_flag_of_all_groups():
    if not os.path.isdir('data/groups/'):
        print('No stored groups data')
        sys.exit()
    ls_res = os.listdir('data/groups/')
    for group_id in ls_res:
        g = Group(group_id)
        g.open()
        g.is_relevant = False


# Write 'False' to field 'is_relevant' of all stored users
def drop_relevance_flag_of_all_users():
    if not os.path.isdir('data/users/'):
        print('No stored users data')
        sys.exit()
    ls_res = os.listdir('data/users/')
    for user_id in ls_res:
        u = User(user_id)
        u.read_info()
        u.is_relevant = False



#! /usr/bin/env python3

import os
import re
import shutil
import glob
from setuptools import setup, find_packages, Command


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""

    CLEAN_FILES = ('./build', './dist', './*.pyc', './*.tgz', './*.egg-info')
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        root = os.path.dirname(__file__)
        for path in self.CLEAN_FILES:
            path = os.path.normpath(os.path.join(root, path))
            for path in glob.glob(path):
                print('removing {}'.format(os.path.relpath(path)))
                shutil.rmtree(path, ignore_errors=True)


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


def read_reqs(fname):
    with open(fname, 'r') as f:
        lines = f.read().split('\n')
        lines = [re.sub(r'[\s]*(#.*)?$', '', x) for x in lines]
        lines = [x for x in lines if x]
    return lines


setup(
    name = 'vkts',
    fullname = 'VK Thematic Search',
    version = '0.0.6',
    author = 'Denis Stepnov',
    author_email = 'stepnovdenis@gmail.com',
    url = 'https://github.com/smurphik/vkts',
    description = ('Library of accelerated access to vk API'
                   ' & thematic students search tool'),
    long_description = read('README.md'),
    long_description_content_type = 'text/markdown',
    license = 'MIT',
    keywords = 'vk social network web scraping',
    packages = find_packages(exclude=['tests']),
    install_requires = read_reqs('requirements.txt'),
    cmdclass = {'clean': CleanCommand},
    entry_points = {'console_scripts': ['vkts = vkts.main:main']},
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'Topic :: Internet :: WWW/HTTP',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
    ],
)



# https://github.com/good-move/VKAuth

import requests
import getpass
from html.parser import HTMLParser


class FormParser(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)
        self.url         = None
        self.denial_url  = None
        self.params      = {}
        self.method      = 'GET'
        self.in_form     = False
        self.in_denial   = False
        self.form_parsed = False

    def handle_starttag(self, tag, attrs):
        tag = tag.lower()
        if tag == 'form':
            if self.in_form:
                raise RuntimeError('Nested form tags are not supported yet')
            else:
                self.in_form = True
        if not self.in_form:
            return

        attrs = dict((name.lower(), value) for name, value in attrs)

        if tag == 'form':
            self.url = attrs['action']
            if 'method' in attrs:
                self.method = attrs['method']
        elif tag == 'input' and 'type' in attrs and 'name' in attrs:
            if attrs['type'] in ['hidden', 'text', 'password']:
                self.params[attrs['name']] = \
                    attrs['value'] if 'value' in attrs else ''
        elif tag == 'input' and 'type' in attrs:
            if attrs['type'] == 'submit':
                self.params['submit_allow_access'] = True
        elif tag == 'div' and 'class' in attrs:
            if attrs['class'] == 'near_btn':
                self.in_denial = True
        elif tag == 'a' and 'href' in attrs and self.in_denial:
            self.denial_url = attrs['href']

    def handle_endtag(self, tag):
        tag = tag.lower()
        if tag == 'form':
            if not self.in_form:
                raise RuntimeError('Unexpected end of <form>')
            self.form_parsed = True
            self.in_form = False
        elif tag == 'div' and self.in_denial:
            self.in_denial = False


class VKAuth(object):

    def __init__(self, permissions, app_id, api_v, email=None, pswd=None,
                 two_factor_auth=False, security_code=None, auto_access=True):
        """
        @args:
            permissions: list of Strings with permissions to get from API
            app_id: (String) vk app id that one can get from vk.com
            api_v: (String) vk API version
        """

        self.session         = requests.Session()
        self.form_parser     = FormParser()
        self._user_id        = None
        self._access_token   = None
        self.response        = None

        self.permissions     = permissions
        self.api_v           = api_v
        self.app_id          = app_id
        self.two_factor_auth = two_factor_auth
        self.security_code   = security_code
        self.email           = email
        self.pswd            = pswd
        self.auto_access     = auto_access

        if security_code is not None and not two_factor_auth:
            raise RuntimeError(
                'Security code provided for non-two-factor authorization')

    def auth(self):
        """
            1. Asks vk.com for app authentication for a user
            2. If user isn't logged in, asks for email and password
            3. Retreives access token and user id
        """
        api_auth_url = 'https://oauth.vk.com/authorize'
        app_id = self.app_id
        permissions = self.permissions
        redirect_uri = 'https://oauth.vk.com/blank.html'
        display = 'wap'
        api_version = self.api_v

        auth_url_template = ('{0}?client_id={1}&scope={2}&redirect_uri={3}'
                             '&display={4}&v={5}&response_type=token')
        auth_url = auth_url_template.format(api_auth_url, app_id,
                                            ','.join(permissions),
                                            redirect_uri, display, api_version)

        self.response = self.session.get(auth_url)

        # look for <form> element in response html and parse it
        if not self._parse_form():
            raise RuntimeError(
                'No <form> element found. Please, check url address')
        else:
            # try to log in with email and password
            # (stored or expected to be entered)
            if not self._log_in():
                return

            # handling two-factor authentication
            # expecting a security code to enter here
            if self.two_factor_auth:
                self._two_fact_auth()

            # allow vk to use this app and access self.permissions
            self._allow_access()

            # now get _access_token and _user_id
            self._get_params()

            # close current session
            self._close()

    def get_token(self):
        """
            @return value:
                None if _access_token == None
                (String) access_token that was retreived in self.auth() method
        """
        return self._access_token

    def get_user_id(self):
        """
            @return value:
                None if _user_id == None
                (String) _user_id that was retreived in self.auth() method
        """
        return self._user_id

    def _parse_form(self):

        self.form_parser = FormParser()
        parser = self.form_parser

        try:
            parser.feed(str(self.response.content))
        except Exception as e:
            print('Unexpected error occured while looking for <form> '
                  'element: ' + str(e))
            return False

        return True

    def _submit_form(self, *params):

        parser = self.form_parser

        if parser.method == 'post':
            payload = parser.params
            payload.update(*params)
            try:
                self.response = self.session.post(parser.url, data=payload)
            except requests.exceptions.RequestException as err:
                print("Error: ", err)
            except requests.exceptions.HTTPError as err:
                print("Error: ", err)
            except requests.exceptions.ConnectionError as err:
                print("Error: ConnectionError\n", err)
            except requests.exceptions.Timeout as err:
                print("Error: Timeout\n", err)
            except Exception as e:
                print("Unexpecred error occured: " + str(e))

        else:
            self.response = None

    def _log_in(self):

        if self.email is None:
            self.email = ''
            while self.email.strip() == '':
                self.email = input('Enter an email to log in: ')

        if self.pswd is None:
            self.pswd = ''
            while self.pswd.strip() == '':
                self.pswd = getpass.getpass('Enter the password: ')

        self._submit_form({'email': self.email, 'pass': self.pswd})

        if not self._parse_form():
            raise RuntimeError(
                    'No <form> element found. Please, check url address')

        # if wrong email or password
        if 'pass' in self.form_parser.params:
            print('Wrong email or password')
            self.email = None
            self.pswd = None
            return False
        elif 'code' in self.form_parser.params and not self.two_factor_auth:
            self.two_factor_auth = True
        else:
            return True

    def _two_fact_auth(self):

        prefix = 'https://m.vk.com'

        if prefix not in self.form_parser.url:
            self.form_parser.url = prefix + self.form_parser.url

        if self.security_code is None:
            self.security_code = input(
                'Enter security code for two-factor authentication: ')

        self._submit_form({'code': self.security_code})

        if not self._parse_form():
            raise RuntimeError(
                'No <form> element found. Please, check url address')

    def _allow_access(self):

        parser = self.form_parser

        if ('submit_allow_access' in parser.params
                and 'grant_access' in parser.url):

            if not self.auto_access:
                answer = ''
                msg = ('Application needs access to the following details in '
                       'your profile:\n' + str(self.permissions) + '\n'
                       'Allow it to use them? (yes or no)')

                attempts = 5
                while answer not in ['yes', 'no'] and attempts > 0:
                    answer = input(msg).lower().strip()
                    attempts -= 1

                if answer == 'no' or attempts == 0:
                    self.form_parser.url = self.form_parser.denial_url
                    print('Access denied')

            self._submit_form({})

    def _get_params(self):

        # Typical error
        if self.response.url == 'https://oauth.vk.com/error?err=2':
            print('Security error. Perhaps the error disappears if'
                  + ' you relogin in the browser.')

        try:
            params = self.response.url.split('#')[1].split('&')
            self._access_token = params[0].split('=')[1]
            self._user_id = params[2].split('=')[1]
        except IndexError as err:
            print('Coudln\'t fetch token and user id from url:')
            print(self.response.url + '\n')

    def _close(self):
        self.session.close()
        self.response = None
        self.form_parser = None
        self.security_code = None
        self.email = None
        self.pswd = None
